"""
Standardized signals definitions for events within the architecture subdomain ``learning``.

All signals defined in this module must follow the name and versioning
conventions specified in docs/decisions/0002-events-naming-and-versioning.rst

They also must comply with the payload definition specified in
docs/decisions/0003-events-payload.rst
"""

from openedx_events.learning.data import (
    BadgeData,
    CcxCoursePassingStatusData,
    CertificateData,
    CohortData,
    CourseAccessRoleData,
    CourseDiscussionConfigurationData,
    CourseEnrollmentData,
    CourseNotificationData,
    CoursePassingStatusData,
    DiscussionThreadData,
    ExamAttemptData,
    ORASubmissionData,
    PersistentCourseGradeData,
    ProgramCertificateData,
    UserData,
    UserNotificationData,
    VerificationAttemptData,
    XBlockSkillVerificationData,
)
from openedx_events.tooling import OpenEdxPublicSignal

# .. event_type: org.openedx.learning.student.registration.completed.v1
# .. event_name: STUDENT_REGISTRATION_COMPLETED
# .. event_key_field: user.pii.username
# .. event_description: Emitted when a user completes registration in Open edX.
# .. event_data: UserData
# .. event_trigger_repository: openedx/edx-platform
STUDENT_REGISTRATION_COMPLETED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.student.registration.completed.v1",
    data={
        "user": UserData,
    }
)


# .. event_type: org.openedx.learning.auth.session.login.completed.v1
# .. event_name: SESSION_LOGIN_COMPLETED
# .. event_key_field: user.pii.username
# .. event_description: Emitted when a user logs in to Open edX.
# .. event_data: UserData
# .. event_trigger_repository: openedx/edx-platform
SESSION_LOGIN_COMPLETED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.auth.session.login.completed.v1",
    data={
        "user": UserData,
    }
)


# .. event_type: org.openedx.learning.course.enrollment.created.v1
# .. event_name: COURSE_ENROLLMENT_CREATED
# .. event_key_field: enrollment.course.course_key
# .. event_description: Emitted when the user enrolls in a course.
# .. event_data: CourseEnrollmentData
# .. event_trigger_repository: openedx/edx-platform
COURSE_ENROLLMENT_CREATED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.course.enrollment.created.v1",
    data={
        "enrollment": CourseEnrollmentData,
    }
)


# .. event_type: org.openedx.learning.course.enrollment.changed.v1
# .. event_name: COURSE_ENROLLMENT_CHANGED
# .. event_key_field: enrollment.course.course_key
# .. event_description: Emitted when the enrollment for a user in a course changes.
# .. event_data: CourseEnrollmentData
# .. event_trigger_repository: openedx/edx-platform
COURSE_ENROLLMENT_CHANGED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.course.enrollment.changed.v1",
    data={
        "enrollment": CourseEnrollmentData,
    }
)


# .. event_type: org.openedx.learning.course.unenrollment.completed.v1
# .. event_name: COURSE_UNENROLLMENT_COMPLETED
# .. event_key_field: enrollment.course.course_key
# .. event_description: Emitted when the user unenrolls from a course.
# .. event_data: CourseEnrollmentData
# .. event_trigger_repository: openedx/edx-platform
COURSE_UNENROLLMENT_COMPLETED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.course.unenrollment.completed.v1",
    data={
        "enrollment": CourseEnrollmentData,
    }
)


# .. event_type: org.openedx.learning.certificate.created.v1
# .. event_name: CERTIFICATE_CREATED
# .. event_key_field: certificate.course.course_key
# .. event_description: Emitted when a certificate is created for a user.
# .. event_data: CertificateData
# .. event_trigger_repository: openedx/edx-platform
CERTIFICATE_CREATED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.certificate.created.v1",
    data={
        "certificate": CertificateData,
    }
)

# .. event_type: org.openedx.learning.program.certificate.awarded.v1
# .. event_name: PROGRAM_CERTIFICATE_AWARDED
# .. event_key_field: program_certificate.program.uuid
# .. event_description: Emitted when a program certificate is awarded to a learner.
# .. event_data: ProgramCertificateData
# .. event_trigger_repository: openedx/credentials
PROGRAM_CERTIFICATE_AWARDED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.program.certificate.awarded.v1",
    data={
        "program_certificate": ProgramCertificateData,
    }
)

# .. event_type: org.openedx.learning.certificate.changed.v1
# .. event_name: CERTIFICATE_CHANGED
# .. event_description: Emitted when the user's certificate changes.
# .. event_data: CertificateData
# .. event_trigger_repository: openedx/edx-platform
CERTIFICATE_CHANGED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.certificate.changed.v1",
    data={
        "certificate": CertificateData,
    }
)


# .. event_type: org.openedx.learning.certificate.revoked.v1
# .. event_name: CERTIFICATE_REVOKED
# .. event_key_field: certificate.course.course_key
# .. event_description: Emitted when a certificate is revoked from a user.
# .. event_data: CertificateData
# .. event_trigger_repository: openedx/edx-platform
CERTIFICATE_REVOKED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.certificate.revoked.v1",
    data={
        "certificate": CertificateData,
    }
)

# .. event_type: org.openedx.learning.program.certificate.revoked.v1
# .. event_name: PROGRAM_CERTIFICATE_REVOKED
# .. event_key_field: program_certificate.program.uuid
# .. event_description: Emit when a program certificate is revoked from a learner.
# .. event_data: ProgramCertificateData
# .. event_trigger_repository: openedx/credentials
PROGRAM_CERTIFICATE_REVOKED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.program.certificate.revoked.v1",
    data={
        "program_certificate": ProgramCertificateData,
    }
)

# .. event_type: org.openedx.learning.cohort_membership.changed.v1
# .. event_name: COHORT_MEMBERSHIP_CHANGED
# .. event_description: Emitted when a user's cohort membership changes.
# .. event_data: CohortData
# .. event_trigger_repository: openedx/edx-platform
COHORT_MEMBERSHIP_CHANGED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.cohort_membership.changed.v1",
    data={
        "cohort": CohortData,
    }
)


# .. event_type: org.openedx.learning.discussions.configuration.changed.v1
# .. event_name: COURSE_DISCUSSIONS_CHANGED
# .. event_description: Emitted when the configuration for a course's discussions changes in the course.
# .. event_warning: This event is currently incompatible with the event bus, list with complex types
#      like ``List[DiscussionTopicContext]`` cannot be serialized yet.
# .. event_data: CourseDiscussionConfigurationData
# .. event_trigger_repository: openedx/edx-platform
COURSE_DISCUSSIONS_CHANGED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.discussions.configuration.changed.v1",
    data={
        "configuration": CourseDiscussionConfigurationData
    }
)

# .. event_type: org.openedx.learning.course.persistent_grade_summary.changed.v1
# .. event_name: PERSISTENT_GRADE_SUMMARY_CHANGED
# .. event_description: Emitted when a course's persistent grade summary changes for a user.
# .. event_data: PersistentCourseGradeData
# .. event_trigger_repository: openedx/edx-platform
PERSISTENT_GRADE_SUMMARY_CHANGED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.course.persistent_grade_summary.changed.v1",
    data={
        "grade": PersistentCourseGradeData,
    }
)


# .. event_type: org.openedx.learning.xblock.skill.verified.v1
# .. event_name: XBLOCK_SKILL_VERIFIED
# .. event_key_field: xblock_info.usage_key
# .. event_description: Emitted when an XBlock skill is verified.
# .. event_data: XBlockSkillVerificationData
# .. event_trigger_repository: openedx/xblock-skill-tagging
XBLOCK_SKILL_VERIFIED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.xblock.skill.verified.v1",
    data={
        "xblock_info": XBlockSkillVerificationData,
    }
)

# .. event_type: org.openedx.learning.user.notification.requested.v1
# .. event_name: USER_NOTIFICATION_REQUESTED
# .. event_description: Can be emitted from apps to send user notifications.
# .. event_data: UserNotificationData
# .. event_warning: This event is currently incompatible with the event bus, list without annotations
#      are not serializable.
# .. event_trigger_repository: openedx/edx-platform openedx/edx-ora2
USER_NOTIFICATION_REQUESTED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.user.notification.requested.v1",
    data={
        "notification_data": UserNotificationData,
    }
)

# .. event_type: org.openedx.learning.exam.attempt.submitted.v1
# .. event_name: EXAM_ATTEMPT_SUBMITTED
# .. event_description: Emitted when an exam attempt is submitted by a learner in edx-exams.
# .. event_data: ExamAttemptData
# .. event_trigger_repository: edx/edx-exams
EXAM_ATTEMPT_SUBMITTED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.exam.attempt.submitted.v1",
    data={
        "exam_attempt": ExamAttemptData,
    }
)

# .. event_type: org.openedx.learning.exam.attempt.rejected.v1
# .. event_name: EXAM_ATTEMPT_REJECTED
# .. event_description: Emitted when an exam attempt is marked rejected in edx-exams.
# .. event_data: ExamAttemptData
# .. event_trigger_repository: edx/edx-exams
EXAM_ATTEMPT_REJECTED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.exam.attempt.rejected.v1",
    data={
        "exam_attempt": ExamAttemptData,
    }
)

# .. event_type: org.openedx.learning.exam.attempt.verified.v1
# .. event_name: EXAM_ATTEMPT_VERIFIED
# .. event_description: Emitted when an exam attempt is marked verified in edx-exams.
# .. event_data: ExamAttemptData
# .. event_trigger_repository: edx/edx-exams
EXAM_ATTEMPT_VERIFIED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.exam.attempt.verified.v1",
    data={
        "exam_attempt": ExamAttemptData,
    }
)

# .. event_type: org.openedx.learning.exam.attempt.errored.v1
# .. event_name: EXAM_ATTEMPT_ERRORED
# .. event_description: Emitted when a learner's exam attempt errors out in edx-exams.
# .. event_data: ExamAttemptData
# .. event_trigger_repository: edx/edx-exams
EXAM_ATTEMPT_ERRORED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.exam.attempt.errored.v1",
    data={
        "exam_attempt": ExamAttemptData,
    }
)

# .. event_type: org.openedx.learning.exam.attempt.reset.v1
# .. event_name: EXAM_ATTEMPT_RESET
# .. event_description: Emitted when an exam attempt is reset in edx-exams.
# .. event_data: ExamAttemptData
# .. event_trigger_repository: edx/edx-exams
EXAM_ATTEMPT_RESET = OpenEdxPublicSignal(
    event_type="org.openedx.learning.exam.attempt.reset.v1",
    data={
        "exam_attempt": ExamAttemptData,
    }
)

# .. event_type: org.openedx.learning.user.course_access_role.added.v1
# .. event_name: COURSE_ACCESS_ROLE_ADDED
# .. event_key_field: course_access_role_data.course_key
# .. event_description: Emitted when a user is given a course access role.
# .. event_data: CourseAccessRoleData
COURSE_ACCESS_ROLE_ADDED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.user.course_access_role.added.v1",
    data={
        "course_access_role_data": CourseAccessRoleData,
    }
)

# .. event_type: org.openedx.learning.user.course_access_role.removed.v1
# .. event_name: COURSE_ACCESS_ROLE_REMOVED
# .. event_key_field: course_access_role_data.course_key
# .. event_description: Emitted when a course access role is removed from a user.
# .. event_data: CourseAccessRoleData
# .. event_trigger_repository: openedx/edx-platform
COURSE_ACCESS_ROLE_REMOVED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.user.course_access_role.removed.v1",
    data={
        "course_access_role_data": CourseAccessRoleData,
    }
)

# .. event_type: org.openedx.learning.forum.thread.created.v1
# .. event_name: FORUM_THREAD_CREATED
# .. event_description: Emitted when a new thread is created in a discussion.
# .. event_data: DiscussionThreadData
# .. event_trigger_repository: openedx/edx-platform
# .. event_warning: This event is currently incompatible with the event bus, dict without annotations
#      are not serializable.
FORUM_THREAD_CREATED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.forum.thread.created.v1",
    data={
        "thread": DiscussionThreadData,
    }
)

# .. event_type: org.openedx.learning.forum.thread.response.created.v1
# .. event_name: FORUM_THREAD_RESPONSE_CREATED
# .. event_description: Emitted when a new response is added to a thread.
# .. event_data: DiscussionThreadData
# .. event_trigger_repository: openedx/edx-platform
# .. event_warning: This event is currently incompatible with the event bus, dict without annotations
#      are not serializable.
FORUM_THREAD_RESPONSE_CREATED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.forum.thread.response.created.v1",
    data={
        "thread": DiscussionThreadData,
    }
)

# .. event_type: org.openedx.learning.forum.thread.response.comment.created.v1
# .. event_name: FORUM_RESPONSE_COMMENT_CREATED
# .. event_description: Emitted when a new comment is added to a response.
# .. event_data: DiscussionThreadData
# .. event_trigger_repository: openedx/edx-platform
# .. event_warning: This event is currently incompatible with the event bus, dict without annotations
#      are not serializable.
FORUM_RESPONSE_COMMENT_CREATED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.forum.thread.response.comment.created.v1",
    data={
        "thread": DiscussionThreadData,
    }
)


# .. event_type: org.openedx.learning.course.notification.requested.v1
# .. event_name: COURSE_NOTIFICATION_REQUESTED
# .. event_description: Emitted when a notification is requested for a course.
# .. event_data: CourseNotificationData
# .. event_trigger_repository: openedx/edx-platform
# .. event_warning: This event is currently incompatible with the event bus, dict without annotations
#      are not serializable.
COURSE_NOTIFICATION_REQUESTED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.course.notification.requested.v1",
    data={
        "course_notification_data": CourseNotificationData,
    }
)


# .. event_type: org.openedx.learning.ora.submission.created.v1
# .. event_name: ORA_SUBMISSION_CREATED
# .. event_description: Emitted when a user submits an ORA assignment.
# .. event_data: ORASubmissionData
# .. event_trigger_repository: openedx/edx-ora2
# .. event_warning: This event is currently incompatible with the event bus, list[dict] cannot be serialized yet.
ORA_SUBMISSION_CREATED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.ora.submission.created.v1",
    data={
        "submission": ORASubmissionData,
    },
)


# .. event_type: org.openedx.learning.course.passing.status.updated.v1
# .. event_name: COURSE_PASSING_STATUS_UPDATED
# .. event_description: Emitted when course grade updates.
# .. event_data: CoursePassingStatusData
# .. event_trigger_repository: openedx/edx-platform
COURSE_PASSING_STATUS_UPDATED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.course.passing.status.updated.v1",
    data={
        "course_passing_status": CoursePassingStatusData,
    }
)


# .. event_type: org.openedx.learning.ccx.course.passing.status.updated.v1
# .. event_name: CCX_COURSE_PASSING_STATUS_UPDATED
# .. event_description: Emitted when a CCX course grade updates.
# .. event_data: CcxCoursePassingStatusData
# .. event_trigger_repository: openedx/edx-platform
CCX_COURSE_PASSING_STATUS_UPDATED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.ccx.course.passing.status.updated.v1",
    data={
        "course_passing_status": CcxCoursePassingStatusData,
    }
)


# .. event_type: org.openedx.learning.badge.awarded.v1
# .. event_name: BADGE_AWARDED
# .. event_description: Emit when a badge is awarded to a learner.
# .. event_data: BadgeData
# .. event_trigger_repository: openedx/credentials
BADGE_AWARDED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.badge.awarded.v1",
    data={
        "badge": BadgeData,
    }
)


# .. event_type: org.openedx.learning.badge.revoked.v1
# .. event_name: BADGE_REVOKED
# .. event_description: Emit when a badge is revoked for a learner,
# .. event_data: BadgeData
# .. event_trigger_repository: openedx/credentials
BADGE_REVOKED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.badge.revoked.v1",
    data={
        "badge": BadgeData,
    }
)


# .. event_type: org.openedx.learning.idv_attempt.created.v1
# .. event_name: IDV_ATTEMPT_CREATED
# .. event_description: Emitted when an IDV attempt is created.
# .. event_data: VerificationAttemptData
# .. event_trigger_repository: openedx/edx-platform
IDV_ATTEMPT_CREATED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.idv_attempt.created.v1",
    data={
        "idv_attempt": VerificationAttemptData,
    }
)


# .. event_type: org.openedx.learning.idv_attempt.pending.v1
# .. event_name: IDV_ATTEMPT_PENDING
# .. event_description: Emitted when an IDV attempt is marked as pending.
# .. event_data: VerificationAttemptData
# .. event_trigger_repository: openedx/edx-platform
IDV_ATTEMPT_PENDING = OpenEdxPublicSignal(
    event_type="org.openedx.learning.idv_attempt.pending.v1",
    data={
        "idv_attempt": VerificationAttemptData,
    }
)


# .. event_type: org.openedx.learning.idv_attempt.approved.v1
# .. event_name: IDV_ATTEMPT_APPROVED
# .. event_description: Emitted when an IDV attempt is approved.
# .. event_data: VerificationAttemptData
# .. event_trigger_repository: openedx/edx-platform
IDV_ATTEMPT_APPROVED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.idv_attempt.approved.v1",
    data={
        "idv_attempt": VerificationAttemptData,
    }
)


# .. event_type: org.openedx.learning.idv_attempt.denied.v1
# .. event_name: IDV_ATTEMPT_DENIED
# .. event_description: Emitted when an IDV attempt is denied.
# .. event_data: VerificationAttemptData
# .. event_trigger_repository: openedx/edx-platform
IDV_ATTEMPT_DENIED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.idv_attempt.denied.v1",
    data={
        "idv_attempt": VerificationAttemptData,
    }
)



"""
Data attributes for events within the architecture subdomain ``enterprise``.

These attributes follow the form of attr objects specified in OEP-49 data
pattern.
"""
from datetime import datetime
from uuid import UUID

import attr
from opaque_keys.edx.keys import CourseKey


@attr.s(frozen=True)
class SubsidyRedemption:
    """
    Data related to a Subsidy Redemption object.

    Attributes:
        subsidy_identifier (str): unique identifier to fetch the applied subsidy
        content_key (str): content id where subsidy is utilized
        lms_user_id (str): lms user id of subsidy beneficiary
    """

    subsidy_identifier = attr.ib(type=str)
    content_key = attr.ib(type=str)
    lms_user_id = attr.ib(type=int)


@attr.s(frozen=True)
class BaseLedgerTransaction:
    """
    Data related to a Ledger Transaction object.

    Attributes:
        uuid (UUID): Primary identifier of the record.
        created (datetime): When the record was created.
        modified (datetime): When the record was last modified.
        idempotency_key (str): Client-generated unique value to achieve idempotency of operations.
        quantity (int): How many units of value this transaction represents (e.g. USD cents).
        state (str): Current lifecyle state of the record, one of (created, pending, committed, failed).
    """

    uuid = attr.ib(type=UUID)
    created = attr.ib(type=datetime)
    modified = attr.ib(type=datetime)
    idempotency_key = attr.ib(type=str)
    quantity = attr.ib(type=int)
    state = attr.ib(type=str)


@attr.s(frozen=True)
class LedgerTransactionReversal(BaseLedgerTransaction):
    """
    Attributes of an ``openedx_ledger.Reversal`` record.

    A ``Reversal`` is a model that represents the "undo-ing" of a ``Transaction`` (see below). It's primarily
    used within the domain of edX Enterprise for recording unenrollments and refunds of subsidized
    enterprise enrollments.
    https://github.com/openedx/openedx-ledger/blob/master/openedx_ledger/models.py

    Attributes:
        uuid (str): Primary identifier of the record.
        created (datetime): When the record was created.
        modified (datetime): When the record was last modified.
        idempotency_key (str): Client-generated unique value to achieve idempotency of operations.
        quantity (int): How many units of value this reversal represents (e.g. USD cents).
        state (str): Current lifecyle state of the record, one of (created, pending, committed, failed).
    """


@attr.s(frozen=True)
class LedgerTransaction(BaseLedgerTransaction):
    """
    Attributes of an ``openedx_ledger.Transaction`` record.

    A ``Transaction`` is a model that represents value moving in or out of a ``Ledger``. It's primarily
    used within the domain of edX Enterprise for recording the redemption of subsidized enrollments.
    https://github.com/openedx/openedx-ledger/blob/master/openedx_ledger/models.py

    Attributes:
        uuid (UUID): Primary identifier of the Transaction.
        created (datetime): When the record was created.
        modified (datetime): When the record was last modified.
        idempotency_key (str): Client-generated unique value to achieve idempotency of operations.
        quantity (int): How many units of value this transaction represents (e.g. USD cents).
        state (str): Current lifecyle state of the record, one of (created, pending, committed, failed).
        ledger_uuid (UUID): The primary identifier of this Transaction's ledger object.
        subsidy_access_policy_uuid (UUID): The primary identifier of the subsidy access policy for this transaction.
        lms_user_id (int): The LMS user id of the user associated with this transaction.
        content_key (CourseKey): The course (run) key associated with this transaction.
        parent_content_key (str): The parent (just course, not run) key for the course key.
        fulfillment_identifier (str): The identifier of the subsidized enrollment record for a learner,
           generated durning enrollment.
        reversal (LedgerTransactionReversal): Any reversal associated with this transaction.
    """

    ledger_uuid = attr.ib(type=UUID)
    subsidy_access_policy_uuid = attr.ib(type=UUID)
    lms_user_id = attr.ib(type=int)
    content_key = attr.ib(type=CourseKey)
    parent_content_key = attr.ib(type=str, default=None)
    fulfillment_identifier = attr.ib(type=str, default=None)
    reversal = attr.ib(type=LedgerTransactionReversal, default=None)


@attr.s(frozen=True)
class EnterpriseCustomerUser:
    """
    Data related to an Enterprise Customer User object.

    Django model definition: https://github.com/openedx/edx-enterprise/blob/cc873d6/enterprise/models.py#L1036

    Attributes:
        id (int): Primary identifier of the record.
        created (datetime): When the record was created.
        modified (datetime): When the record was last modified.
        enterprise_customer_uuid (UUID): The enterprise customer to which the user is linked.
        user_id (int): The LMS user ID corresponding to this enterprise user.
        active (bool): The active enterprise user for the given LMS user.
        linked (bool): This enterprise user has been linked to an enterprise customer.
        is_relinkable (bool): When set to False, the user cannot be relinked to the enterprise.
        invite_key (UUID): Invite key used to link a learner to an enterprise.
        should_inactivate_other_customers (bool): When enabled along with `active`, all other linked enterprise
           customers for this user will be marked as inactive upon save.
    """

    id = attr.ib(type=int)
    created = attr.ib(type=datetime)
    modified = attr.ib(type=datetime)
    enterprise_customer_uuid = attr.ib(type=UUID)
    user_id = attr.ib(type=int)
    active = attr.ib(type=bool)
    linked = attr.ib(type=bool)
    is_relinkable = attr.ib(type=bool)
    should_inactivate_other_customers = attr.ib(type=bool)
    invite_key = attr.ib(type=UUID, default=None)


@attr.s(frozen=True)
class EnterpriseCourseEnrollment:
    """
    Data related to an Enterprise Course Enrollment object.

    Django model definition: https://github.com/openedx/edx-enterprise/blob/cc873d6/enterprise/models.py#L1983

    Attributes:
        id (int): Primary identifier of the record.
        created (datetime): When the record was created.
        modified (datetime): When the record was last modified.
        enterprise_customer_user (EnterpriseCustomerUser): The enterprise learner to which this enrollment is attached.
        course_id (CourseKey): The ID of the course in which the learner was enrolled.
        saved_for_later (bool): Specifies whether a user marked this course as saved for later in the learner portal.
        source_slug (str): DB slug for the source of the enrollment, e.g. "enrollment_task", "management_command", etc.
        unenrolled (bool): Specifies whether the related LMS course enrollment object was unenrolled.
        unenrolled_at (datetime): Specifies when the related LMS course enrollment object was unenrolled.
    """

    id = attr.ib(type=int)
    created = attr.ib(type=datetime)
    modified = attr.ib(type=datetime)
    enterprise_customer_user = attr.ib(type=EnterpriseCustomerUser)
    course_id = attr.ib(type=CourseKey)
    saved_for_later = attr.ib(type=bool)
    source_slug = attr.ib(type=str, default=None)
    unenrolled = attr.ib(type=bool, default=None)
    unenrolled_at = attr.ib(type=datetime, default=None)


@attr.s(frozen=True)
class BaseEnterpriseFulfillment:
    """
    Defines the common attributes of enterprise fulfillment classes, i.e. ``enterprise.EnterpriseFulfillmentSource``.

    Django model definition: https://github.com/openedx/edx-enterprise/blob/cc873d6/enterprise/models.py#L2213

    Attributes:
        uuid (str): Primary identifier of the record.
        created (datetime): When the record was created.
        modified (datetime): When the record was last modified.
        fulfillment_type (str): Subsidy fulfillment type, typical values: "license", "learner_credit", "coupon_code".
        enterprise_course_entitlement_uuid (UUID): The course entitlement the associated subsidy is for.
        enterprise_course_enrollment (EnterpriseCourseEnrollment): The course enrollment the associated subsidy is
           for.
        is_revoked (bool): Whether the enterprise subsidy is revoked, e.g., when a user's license is revoked.
    """

    uuid = attr.ib(type=UUID)
    created = attr.ib(type=datetime)
    modified = attr.ib(type=datetime)
    fulfillment_type = attr.ib(type=str)
    is_revoked = attr.ib(type=bool)
    enterprise_course_entitlement_uuid = attr.ib(type=UUID, default=None)
    enterprise_course_enrollment = attr.ib(type=EnterpriseCourseEnrollment, default=None)


@attr.s(frozen=True)
class LearnerCreditEnterpriseCourseEnrollment(BaseEnterpriseFulfillment):
    """
    Attributes of an ``enterprise.LearnerCreditEnterpriseCourseEnrollment`` record.

    Django model definition: https://github.com/openedx/edx-enterprise/blob/cc873d6/enterprise/models.py#L2325

    All of the same attributes from BaseEnterpriseFulfillment plus the following:

    Attributes:
        transaction_id (UUID): Ledgered transaction UUID to associate with this learner credit fulfillment.
    """

    transaction_id = attr.ib(type=UUID, default=None)


@attr.s(frozen=True)
class LicensedEnterpriseCourseEnrollment(BaseEnterpriseFulfillment):
    """
    Attributes of an ``enterprise.LicensedEnterpriseCourseEnrollment`` record.

    Django model definition: https://github.com/openedx/edx-enterprise/blob/cc873d6/enterprise/models.py#L2355

    All of the same attributes from BaseEnterpriseFulfillment plus the following:

    Attributes:
        license_uuid (UUID): License UUID to associate with this enterprise license fulfillment.
    """

    license_uuid = attr.ib(type=UUID, default=None)



"""Test custom servializers"""
from unittest import TestCase
from uuid import UUID, uuid4

from ccx_keys.locator import CCXLocator

from openedx_events.event_bus.avro.custom_serializers import CcxCourseLocatorAvroSerializer, UuidAvroSerializer


class TestCCXLocatorSerailizer(TestCase):
    """Test case for CCXLocator serializer."""

    def test_serialize(self):
        """
        Test case for serializing CCXLocator object.
        """

        obj1 = CCXLocator(org="edx", course="DemoX", run="Demo_course", ccx="1")
        expected1 = "ccx-v1:edx+DemoX+Demo_course+ccx@1"
        result1 = CcxCourseLocatorAvroSerializer.serialize(obj1)
        self.assertEqual(result1, expected1)

    def test_deserialize(self):
        """
        Test case for deserializing CCXLocator object.
        """

        data1 = "ccx-v1:edx+DemoX+Demo_course+ccx@1"
        expected1 = CCXLocator(org="edx", course="DemoX", run="Demo_course", ccx="1")
        result1 = CcxCourseLocatorAvroSerializer.deserialize(data1)
        self.assertEqual(result1, expected1)


class TestUuidAvroSerializer(TestCase):
    """
    Tests case for Avro UUID de-/serialization.
    """
    def test_serialize(self):
        """
        Test UUID Avro serialization.
        """
        some_uuid = uuid4()
        expected_result = str(some_uuid)
        actual_result = UuidAvroSerializer.serialize(some_uuid)
        self.assertEqual(actual_result, expected_result)

    def test_deserialize(self):
        """
        Test UUID Avro de-serialization.
        """
        uuid_str = str(uuid4())
        expected_result = UUID(uuid_str)
        actual_result = UuidAvroSerializer.deserialize(uuid_str)
        self.assertEqual(actual_result, expected_result)




"""
Classes to serialize and deserialize custom types used by openedx events. See README for usage.
"""
from abc import ABC, abstractmethod
from datetime import datetime
from uuid import UUID

from ccx_keys.locator import CCXLocator
from opaque_keys.edx.keys import CourseKey, UsageKey
from opaque_keys.edx.locator import LibraryLocatorV2, LibraryUsageLocatorV2

from openedx_events.event_bus.avro.types import PYTHON_TYPE_TO_AVRO_MAPPING


class BaseCustomTypeAvroSerializer(ABC):
    """
    Used by openedx_events.avro_utilities class to serialize/deserialize custom types.
    """

    cls: type
    field_type: str

    @staticmethod
    @abstractmethod
    def serialize(obj) -> str:
        """Abstract method to serialize obj into string."""

    @staticmethod
    @abstractmethod
    def deserialize(data: str) -> object:
        """Abstract method to deserialize string into obj."""


class CourseKeyAvroSerializer(BaseCustomTypeAvroSerializer):
    """
    CustomTypeAvroSerializer for CourseKey class.
    """

    cls = CourseKey
    field_type = PYTHON_TYPE_TO_AVRO_MAPPING[str]

    @staticmethod
    def serialize(obj) -> str:
        """Serialize obj into string."""
        return str(obj)

    @staticmethod
    def deserialize(data: str):
        """Deserialize string into obj."""
        return CourseKey.from_string(data)


class CcxCourseLocatorAvroSerializer(BaseCustomTypeAvroSerializer):
    """
    CustomTypeAvroSerializer for CCXLocator class.
    """

    cls = CCXLocator
    field_type = PYTHON_TYPE_TO_AVRO_MAPPING[str]

    @staticmethod
    def serialize(obj) -> str:
        """Serialize obj into string."""
        return str(obj)

    @staticmethod
    def deserialize(data: str):
        """Deserialize string into obj."""
        return CCXLocator.from_string(data)


class DatetimeAvroSerializer(BaseCustomTypeAvroSerializer):
    """
    CustomTypeAvroSerializer for datetime class.

    Note the choice of an iso-formatted string comes directly from the required CloudEvent <-> Avro mapping
    specified here:
    https://github.com/cloudevents/spec/blob/v1.0.2/cloudevents/formats/avro-format.md#21-type-system-mapping
    """

    cls = datetime
    field_type = PYTHON_TYPE_TO_AVRO_MAPPING[str]

    @staticmethod
    def serialize(obj) -> str:
        """Serialize obj into string."""
        return obj.isoformat()

    @staticmethod
    def deserialize(data: str):
        """Deserialize string into obj."""
        return datetime.fromisoformat(data)


class UsageKeyAvroSerializer(BaseCustomTypeAvroSerializer):
    """
    CustomTypeAvroSerializer for UsageKey class.
    """

    cls = UsageKey
    field_type = PYTHON_TYPE_TO_AVRO_MAPPING[str]

    @staticmethod
    def serialize(obj) -> str:
        """Serialize obj into string."""
        return str(obj)

    @staticmethod
    def deserialize(data: str):
        """Deserialize string into obj."""
        return UsageKey.from_string(data)


class LibraryLocatorV2AvroSerializer(BaseCustomTypeAvroSerializer):
    """
    CustomTypeAvroSerializer for LibraryLocatorV2 class.
    """

    cls = LibraryLocatorV2
    field_type = PYTHON_TYPE_TO_AVRO_MAPPING[str]

    @staticmethod
    def serialize(obj) -> str:
        """Serialize obj into string."""
        return str(obj)

    @staticmethod
    def deserialize(data: str):
        """Deserialize string into obj."""
        return LibraryLocatorV2.from_string(data)


class LibraryUsageLocatorV2AvroSerializer(BaseCustomTypeAvroSerializer):
    """
    CustomTypeAvroSerializer for LibraryUsageLocatorV2 class.
    """

    cls = LibraryUsageLocatorV2
    field_type = PYTHON_TYPE_TO_AVRO_MAPPING[str]

    @staticmethod
    def serialize(obj) -> str:
        """Serialize obj into string."""
        return str(obj)

    @staticmethod
    def deserialize(data: str):
        """Deserialize string into obj."""
        return LibraryUsageLocatorV2.from_string(data)


class UuidAvroSerializer(BaseCustomTypeAvroSerializer):
    """
    CustomTypeAvroSerializer for the UUID class.

    https://github.com/cloudevents/spec/blob/v1.0.2/cloudevents/formats/avro-format.md#21-type-system-mapping
    """

    cls = UUID
    field_type = PYTHON_TYPE_TO_AVRO_MAPPING[str]

    @staticmethod
    def serialize(obj) -> str:
        """Serialize obj into string."""
        return str(obj)

    @staticmethod
    def deserialize(data: str):
        """Deserialize string into obj."""
        return UUID(data)


DEFAULT_CUSTOM_SERIALIZERS = [
    CourseKeyAvroSerializer,
    CcxCourseLocatorAvroSerializer,
    DatetimeAvroSerializer,
    LibraryLocatorV2AvroSerializer,
    LibraryUsageLocatorV2AvroSerializer,
    UsageKeyAvroSerializer,
    UuidAvroSerializer,
]



"""
Standardized signals definitions for events within the architecture subdomain ``enterprise``.

All signals defined in this module must follow the name and versioning
conventions specified in OEP-41.

They also must comply with the payload definition specified in
docs/decisions/0003-events-payload.rst
"""

from openedx_events.enterprise.data import LearnerCreditEnterpriseCourseEnrollment, LedgerTransaction, SubsidyRedemption
from openedx_events.tooling import OpenEdxPublicSignal

# .. event_type: org.openedx.enterprise.subsidy.redeemed.v1
# .. event_name: SUBSIDY_REDEEMED
# .. event_description: emitted when an enterprise subsidy is utilized.
# .. event_data: SubsidyRedemption
# .. event_warning: This event is not being currently used in any of the Open edX services. Review
#      https://github.com/openedx/openedx-events/issues/445 for more information about the future of this event.
SUBSIDY_REDEEMED = OpenEdxPublicSignal(
    event_type="org.openedx.enterprise.subsidy.redeemed.v1",
    data={
        "redemption": SubsidyRedemption,
    }
)

# .. event_type: org.openedx.enterprise.subsidy.redemption-reversed.v1
# .. event_name: SUBSIDY_REDEMPTION_REVERSED
# .. event_description: emitted when an enterprise subsidy is reversed.
# .. event_data: SubsidyRedemption
# .. event_warning: This event is not being currently used in any of the Open edX services. Review
#      https://github.com/openedx/openedx-events/issues/445 for more information about the future of this event.
SUBSIDY_REDEMPTION_REVERSED = OpenEdxPublicSignal(
    event_type="org.openedx.enterprise.subsidy.redemption-reversed.v1",
    data={
        "redemption": SubsidyRedemption,
    }
)


# .. event_type: org.openedx.enterprise.subsidy_ledger_transaction.created.v1
# .. event_name: LEDGER_TRANSACTION_CREATED
# .. event_description: emitted when an enterprise ledger transaction is created.
#      See: https://github.com/openedx/openedx-ledger/tree/main/docs/decisions
# .. event_data: LedgerTransaction
# .. event_trigger_repository: openedx/enterprise-subsidy
LEDGER_TRANSACTION_CREATED = OpenEdxPublicSignal(
    event_type="org.openedx.enterprise.subsidy_ledger_transaction.created.v1",
    data={
        "ledger_transaction": LedgerTransaction,
    }
)


# .. event_type: org.openedx.enterprise.subsidy_ledger_transaction.committed.v1
# .. event_name: LEDGER_TRANSACTION_COMMITTED
# .. event_description: emitted when an enterprise ledger transaction is committed.
#      See: https://github.com/openedx/openedx-ledger/tree/main/docs/decisions
# .. event_data: LedgerTransaction
# .. event_trigger_repository: openedx/enterprise-subsidy
LEDGER_TRANSACTION_COMMITTED = OpenEdxPublicSignal(
    event_type="org.openedx.enterprise.subsidy_ledger_transaction.committed.v1",
    data={
        "ledger_transaction": LedgerTransaction,
    }
)


# .. event_type: org.openedx.enterprise.subsidy_ledger_transaction.failed.v1
# .. event_name: LEDGER_TRANSACTION_FAILED
# .. event_description: emitted when an enterprise ledger transaction fails.
#      See: https://github.com/openedx/openedx-ledger/tree/main/docs/decisions
# .. event_data: LedgerTransaction
# .. event_trigger_repository: openedx/enterprise-subsidy
LEDGER_TRANSACTION_FAILED = OpenEdxPublicSignal(
    event_type="org.openedx.enterprise.subsidy_ledger_transaction.failed.v1",
    data={
        "ledger_transaction": LedgerTransaction,
    }
)


# .. event_type: org.openedx.enterprise.subsidy_ledger_transaction.reversed.v1
# .. event_name: LEDGER_TRANSACTION_REVERSED
# .. event_description: emitted when an enterprise ledger transaction is reversed.
#      See: https://github.com/openedx/openedx-ledger/tree/main/docs/decisions
# .. event_data: LedgerTransaction
# .. event_trigger_repository: openedx/enterprise-subsidy
LEDGER_TRANSACTION_REVERSED = OpenEdxPublicSignal(
    event_type="org.openedx.enterprise.subsidy_ledger_transaction.reversed.v1",
    data={
        "ledger_transaction": LedgerTransaction,
    }
)


# .. event_type: org.openedx.enterprise.learner_credit_course_enrollment.revoked.v1
# .. event_name: LEARNER_CREDIT_COURSE_ENROLLMENT_REVOKED
# .. event_description: emitted when a LearnerCreditEnterpriseCourseEnrollment is revoked. This most often happens when
#      an enterprise learner unenrolls from a course which was LC-subsidized.
# .. event_data: LearnerCreditEnterpriseCourseEnrollment
# .. event_trigger_repository: openedx/edx-enterprise
LEARNER_CREDIT_COURSE_ENROLLMENT_REVOKED = OpenEdxPublicSignal(
    event_type="org.openedx.enterprise.learner_credit_course_enrollment.revoked.v1",
    data={
        "learner_credit_course_enrollment": LearnerCreditEnterpriseCourseEnrollment,
    }
)



"""Tests for avro.deserializer"""
import json
from datetime import datetime
from typing import Dict, List
from unittest import TestCase

import ddt
from opaque_keys.edx.keys import CourseKey, UsageKey
from opaque_keys.edx.locator import LibraryLocatorV2, LibraryUsageLocatorV2

from openedx_events.event_bus.avro.deserializer import AvroSignalDeserializer, deserialize_bytes_to_event_data
from openedx_events.event_bus.avro.tests.test_utilities import (
    ComplexAttrs,
    EventData,
    NestedAttrsWithDefaults,
    NestedNonAttrs,
    NonAttrs,
    SimpleAttrs,
    SimpleAttrsWithDefaults,
    SpecialDeserializer,
    SubTestData0,
    SubTestData1,
    create_simple_signal,
)
from openedx_events.tests.utils import FreezeSignalCacheMixin


@ddt.ddt
class TestAvroSignalDeserializerCache(TestCase, FreezeSignalCacheMixin):
    """Test AvroSignalDeserializer"""

    def setUp(self) -> None:
        super().setUp()
        self.maxDiff = None

    @ddt.data(
        (
            SimpleAttrs,
            {
                "name": "CloudEvent",
                "type": "record",
                "doc": "Avro Event Format for CloudEvents created with openedx_events/schema",
                "namespace": "simple.signal",
                "fields": [
                    {
                        "name": "data",
                        "type": {
                            "name": "SimpleAttrs",
                            "type": "record",
                            "fields": [
                                {"name": "boolean_field", "type": "boolean"},
                                {"name": "int_field", "type": "long"},
                                {"name": "float_field", "type": "double"},
                                {"name": "bytes_field", "type": "bytes"},
                                {"name": "string_field", "type": "string"},
                            ],
                        },
                    },
                ],
            }
        ),
        (
            ComplexAttrs,
            {
                "name": "CloudEvent",
                "type": "record",
                "doc": "Avro Event Format for CloudEvents created with openedx_events/schema",
                "namespace": "simple.signal",
                "fields": [
                    {
                        "name": "data",
                        "type": {
                            "name": "ComplexAttrs",
                            "type": "record",
                            "fields": [
                                {"name": "list_field", "type": {"type": "array", "items": "long"}},
                                {"name": "dict_field", "type": {"type": "map", "values": "long"}},
                            ],
                        },
                    },
                ],
            }
        )
    )
    @ddt.unpack
    def test_schema_string(self, data_cls, expected_schema):
        """
        Test JSON round-trip; schema creation is tested more fully in test_schema.py.
        """
        SIGNAL = create_simple_signal({
            "data": data_cls
        })

        actual_schema = json.loads(AvroSignalDeserializer(SIGNAL).schema_string())

        assert actual_schema == expected_schema

    def test_convert_dict_to_event_data(self):
        SIGNAL = create_simple_signal({"test_data": EventData})
        initial_dict = {
            "test_data": {
                "course_id": "bar.course",
                "sub_name": "foo",
                "sub_test_0": {"course_id": "a.nother.course", "sub_name": "a.sub.name"},
                "sub_test_1": {
                    "course_id": "b.uber.another.course",
                    "sub_name": "b.uber.sub.name",
                },
            }
        }
        deserializer = AvroSignalDeserializer(SIGNAL)
        event_data = deserializer.from_dict(initial_dict)
        expected_event_data = EventData("foo", "bar.course",
                                        SubTestData0("a.sub.name", "a.nother.course"),
                                        SubTestData1("b.uber.sub.name", "b.uber.another.course")
                                        )
        test_data = event_data["test_data"]
        self.assertIsInstance(test_data, EventData)
        self.assertEqual(test_data, expected_event_data)

    def test_default_datetime_deserialization(self):
        SIGNAL = create_simple_signal({"birthday": datetime})
        deserializer = AvroSignalDeserializer(SIGNAL)
        birthday = datetime(year=1989, month=9, day=6)
        as_dict = {"birthday": birthday.isoformat()}
        event_data = deserializer.from_dict(as_dict)
        birthday_deserialized = event_data["birthday"]
        self.assertIsInstance(birthday_deserialized, datetime)
        self.assertEqual(birthday_deserialized, birthday)

    def test_default_coursekey_deserialization(self):
        SIGNAL = create_simple_signal({"course": CourseKey})
        deserializer = AvroSignalDeserializer(SIGNAL)
        course_key = CourseKey.from_string("course-v1:edX+DemoX.1+2014")
        as_dict = {"course": str(course_key)}
        event_data = deserializer.from_dict(as_dict)
        course_deserialized = event_data["course"]
        self.assertIsInstance(course_deserialized, CourseKey)
        self.assertEqual(course_deserialized, course_key)

    def test_default_usagekey_deserialization(self):
        """
        Test deserialization of UsageKey
        """
        SIGNAL = create_simple_signal({"usage_key": UsageKey})
        deserializer = AvroSignalDeserializer(SIGNAL)
        usage_key = UsageKey.from_string(
            "block-v1:edx+DemoX+Demo_course+type@video+block@UaEBjyMjcLW65gaTXggB93WmvoxGAJa0JeHRrDThk",
        )
        as_dict = {"usage_key": str(usage_key)}
        event_data = deserializer.from_dict(as_dict)
        usage_key_deserialized = event_data["usage_key"]
        self.assertIsInstance(usage_key_deserialized, UsageKey)
        self.assertEqual(usage_key_deserialized, usage_key)

    def test_default_librarylocatorv2_deserialization(self):
        """
        Test deserialization of LibraryLocatorV2
        """
        SIGNAL = create_simple_signal({"library_key": LibraryLocatorV2})
        deserializer = AvroSignalDeserializer(SIGNAL)
        library_key = LibraryLocatorV2.from_string("lib:MITx:reallyhardproblems")
        as_dict = {"library_key": str(library_key)}
        event_data = deserializer.from_dict(as_dict)
        library_key_deserialized = event_data["library_key"]
        self.assertIsInstance(library_key_deserialized, LibraryLocatorV2)
        self.assertEqual(library_key_deserialized, library_key)

    def test_default_libraryusagelocatorv2_deserialization(self):
        """
        Test deserialization of LibraryUsageLocatorV2
        """
        SIGNAL = create_simple_signal({"usage_key": LibraryUsageLocatorV2})
        deserializer = AvroSignalDeserializer(SIGNAL)
        usage_key = LibraryUsageLocatorV2.from_string("lb:MITx:reallyhardproblems:problem:problem1")
        as_dict = {"usage_key": str(usage_key)}
        event_data = deserializer.from_dict(as_dict)
        usage_key_deserialized = event_data["usage_key"]
        self.assertIsInstance(usage_key_deserialized, LibraryUsageLocatorV2)
        self.assertEqual(usage_key_deserialized, usage_key)

    def test_deserialization_with_custom_serializer(self):
        SIGNAL = create_simple_signal({"test_data": NonAttrs})
        deserializer = SpecialDeserializer(SIGNAL)
        as_dict = {
            "test_data": "a.val:a.nother.val"
        }
        event_data = deserializer.from_dict(as_dict)
        non_attrs_deserialized = event_data["test_data"]
        self.assertIsInstance(non_attrs_deserialized, NonAttrs)
        self.assertEqual(non_attrs_deserialized, NonAttrs("a.val", "a.nother.val"))

    def test_deserialization_with_custom_serializer_on_nested_fields(self):
        SIGNAL = create_simple_signal({"test_data": NestedNonAttrs})
        as_dict = {"test_data": {
            "field_0": "a.val:a.nother.val"
        }}
        deserializer = SpecialDeserializer(SIGNAL)
        event_data = deserializer.from_dict(as_dict)
        deserialized_nested = event_data["test_data"]
        self.assertIsInstance(deserialized_nested, NestedNonAttrs)
        inner_deserialized_non_attrs = deserialized_nested.field_0
        self.assertIsInstance(inner_deserialized_non_attrs, NonAttrs)
        self.assertEqual(inner_deserialized_non_attrs, NonAttrs("a.val", "a.nother.val"))

    def test_deserialization_fails_if_missing_fields(self):
        # missing "sub_name" field
        initial_dict = {
            "test_data": {
                "course_id": "bar.course",
            }
        }

        SIGNAL = create_simple_signal({"test_data": SubTestData0})
        deserializer = AvroSignalDeserializer(SIGNAL)
        with self.assertRaises(TypeError):
            _ = deserializer.from_dict(initial_dict)

    def test_deserialization_of_optional_fields(self):
        SIGNAL = create_simple_signal({
            "data": SimpleAttrsWithDefaults
        })
        deserializer = AvroSignalDeserializer(SIGNAL)

        as_dict = {"data": {}}
        data_dict = deserializer.from_dict(as_dict)
        self.assertIsInstance(data_dict["data"], SimpleAttrsWithDefaults)
        self.assertEqual(data_dict["data"], SimpleAttrsWithDefaults())

    def test_deserialization_of_nested_optional_fields(self):
        SIGNAL = create_simple_signal({
            "data": NestedAttrsWithDefaults
        })
        deserializer = AvroSignalDeserializer(SIGNAL)
        as_dict = {"data": {"field_0": {}}}
        data_dict = deserializer.from_dict(as_dict)
        nested_field = data_dict["data"].field_0
        self.assertIsInstance(nested_field, SimpleAttrsWithDefaults)
        self.assertEqual(nested_field, SimpleAttrsWithDefaults())

    def test_deserialization_of_list_with_annotation(self):
        """
        Check that deserialization works as expected when list data is annotated.
        """
        LIST_SIGNAL = create_simple_signal({"list_input": List[int]})
        initial_dict = {"list_input": [1, 3]}
        deserializer = AvroSignalDeserializer(LIST_SIGNAL)
        event_data = deserializer.from_dict(initial_dict)
        expected_event_data = [1, 3]
        test_data = event_data["list_input"]
        self.assertIsInstance(test_data, list)
        self.assertEqual(test_data, expected_event_data)

    def test_deserialization_of_list_without_annotation(self):
        """
        Check that deserialization raises error when list data is not annotated.
        """
        # create dummy signal to bypass schema check while initializing deserializer
        # This allows us to test whether correct exceptions are raised while deserializing data
        SIGNAL = create_simple_signal({"list_input": List[int]})
        LIST_SIGNAL = create_simple_signal({"list_input": List})
        initial_dict = {"list_input": [1, 3]}
        deserializer = AvroSignalDeserializer(SIGNAL)
        # Update signal with incomplete type info
        deserializer.signal = LIST_SIGNAL
        with self.assertRaises(TypeError):
            deserializer.from_dict(initial_dict)

    def test_deserialization_of_dict_with_annotation(self):
        """
        Check that deserialization works as expected when dict data is annotated.
        """
        DICT_SIGNAL = create_simple_signal({"dict_input": Dict[str, int]})
        initial_dict = {"dict_input": {"key1": 1, "key2": 3}}

        deserializer = AvroSignalDeserializer(DICT_SIGNAL)
        event_data = deserializer.from_dict(initial_dict)
        expected_event_data = {"key1": 1, "key2": 3}
        test_data = event_data["dict_input"]

        self.assertIsInstance(test_data, dict)
        self.assertEqual(test_data, expected_event_data)

    def test_deserialization_of_dict_without_annotation(self):
        """
        Check that deserialization raises error when dict data is not annotated.

        Create dummy signal to bypass schema check while initializing deserializer. Then,
        update signal with incomplete type info to test whether correct exceptions are raised while deserializing data.
        """
        SIGNAL = create_simple_signal({"dict_input": Dict[str, int]})
        DICT_SIGNAL = create_simple_signal({"dict_input": Dict})
        initial_dict = {"dict_input": {"key1": 1, "key2": 3}}

        deserializer = AvroSignalDeserializer(SIGNAL)
        deserializer.signal = DICT_SIGNAL

        with self.assertRaises(TypeError):
            deserializer.from_dict(initial_dict)

    def test_deserialization_of_dict_with_complex_types_fails(self):
        SIGNAL = create_simple_signal({"dict_input": Dict[str, list]})
        with self.assertRaises(TypeError):
            AvroSignalDeserializer(SIGNAL)
        initial_dict = {"dict_input": {"key1": [1, 3], "key2": [4, 5]}}
        # create dummy signal to bypass schema check while initializing deserializer
        # This allows us to test whether correct exceptions are raised while deserializing data
        DUMMY_SIGNAL = create_simple_signal({"dict_input": Dict[str, int]})
        deserializer = AvroSignalDeserializer(DUMMY_SIGNAL)
        # Update signal with incorrect type info
        deserializer.signal = SIGNAL
        with self.assertRaises(TypeError):
            deserializer.from_dict(initial_dict)

    def test_deserialization_of_dicts_with_keys_of_complex_types_fails(self):
        SIGNAL = create_simple_signal({"dict_input": Dict[CourseKey, int]})
        deserializer = AvroSignalDeserializer(SIGNAL)
        initial_dict = {"dict_input": {CourseKey.from_string("course-v1:edX+DemoX.1+2014"): 1}}
        with self.assertRaises(TypeError):
            deserializer.from_dict(initial_dict)

    def test_deserialization_of_nested_list_fails(self):
        """
        Check that deserialization raises error when nested list data is passed.
        """
        # create dummy signal to bypass schema check while initializing deserializer
        # This allows us to test whether correct exceptions are raised while deserializing data
        SIGNAL = create_simple_signal({"list_input": List[int]})
        LIST_SIGNAL = create_simple_signal({"list_input": List[List[int]]})
        initial_dict = {"list_input": [[1, 3], [4, 5]]}
        deserializer = AvroSignalDeserializer(SIGNAL)
        # Update signal with incomplete type info
        deserializer.signal = LIST_SIGNAL
        with self.assertRaises(TypeError):
            deserializer.from_dict(initial_dict)

    def test_deserialization_of_nested_list_with_complex_types_fails(self):
        SIGNAL = create_simple_signal({"list_input": List[list]})
        with self.assertRaises(TypeError):
            AvroSignalDeserializer(SIGNAL)
        initial_dict = {"list_input": [[1, 3], [4, 5]]}
        # create dummy signal to bypass schema check while initializing deserializer
        # This allows us to test whether correct exceptions are raised while deserializing data
        DUMMY_SIGNAL = create_simple_signal({"list_input": List[int]})
        deserializer = AvroSignalDeserializer(DUMMY_SIGNAL)
        # Update signal with incorrect type info
        deserializer.signal = SIGNAL
        with self.assertRaises(TypeError):
            deserializer.from_dict(initial_dict)

    def test_deserialize_bytes_to_event_data(self):
        """
        Test deserialize_bytes_to_event_data utility function.
        """
        SIGNAL = create_simple_signal({"test_data": EventData})
        bytes_data = b'\x06foo\x14bar.course\x14a.sub.name\x1ea.nother.course\x1eb.uber.sub.name*b.uber.another.course'
        expected = {"test_data": EventData(
            "foo",
            "bar.course",
            SubTestData0("a.sub.name", "a.nother.course"),
            SubTestData1("b.uber.sub.name", "b.uber.another.course"),
        )}
        deserialized = deserialize_bytes_to_event_data(bytes_data, SIGNAL)
        self.assertIsInstance(deserialized["test_data"], EventData)
        self.assertEqual(deserialized, expected)



"""
Data attributes for events within the architecture subdomain `learning`.

These attributes follow the form of attr objects specified in OEP-49 data
pattern.
"""
from datetime import datetime
from typing import List, Optional

import attr
from ccx_keys.locator import CCXLocator
from opaque_keys.edx.keys import CourseKey, UsageKey


@attr.s(frozen=True)
class UserNonPersonalData:
    """
    Data related to a user object that does not contain personal information (PII).

    Attributes:
        id (int): unique identifier for the Django User object.
        is_active (bool): indicates whether the user is active.
    """

    id = attr.ib(type=int)
    is_active = attr.ib(type=bool)


@attr.s(frozen=True)
class UserPersonalData:
    """
    Data related to a user object that contains personal information (PII).

    Attributes:
        username (str): username associated with the user.
        email (str): email associated with the user.
        name (str): name associated with the user's profile.
    """

    username = attr.ib(type=str)
    email = attr.ib(type=str)
    name = attr.ib(type=str, factory=str)


@attr.s(frozen=True)
class UserData(UserNonPersonalData):
    """
    Data related to a user object, including personal information and non-personal information.

    This class extends UserNonPersonalData to include PII data completing the user object.

    Attributes:
        id (int): unique identifier for the Django User object.
        is_active (bool): indicates whether the user is active.
        pii (UserPersonalData): user's Personal Identifiable Information.
    """

    pii = attr.ib(type=UserPersonalData)


@attr.s(frozen=True)
class CourseData:
    """
    Data related to a course object.

    This data is based on the fields available in the CourseOverview data model
    defined in the course_overviews app.

    Attributes:
        course_key (str): identifier of the course.
        display_name (str): display name associated with the course.
        start (datetime): start date for the course. Defaults to None.
        end (datetime): end date for the course. Defaults to None.
    """

    course_key = attr.ib(type=CourseKey)
    display_name = attr.ib(type=str, factory=str)
    start = attr.ib(type=datetime, default=None)
    end = attr.ib(type=datetime, default=None)


@attr.s(frozen=True)
class CcxCourseData:
    """
    Represents data for a CCX (Custom Courses for edX) course.

    Attributes:
        ccx_course_key (CCXLocator): The unique identifier for the CCX course.
        master_course_key (CourseKey): The unique identifier for the original course from which the CCX is derived.
        display_name (str): The name of the CCX course as it should appear to users.
        coach_email (str): The email address of the coach (instructor) for the CCX course.
        start (str, optional): The start date of the CCX course. Defaults to None, indicating no specific start date.
        end (str, optional): The end date of the CCX course. Defaults to None, indicating no specific end date.
        max_students_allowed (int, optional): The maximum number of students that can enroll in the CCX course.
           Defaults to None, indicating no limit.
    """

    ccx_course_key = attr.ib(type=CCXLocator)
    master_course_key = attr.ib(type=CourseKey)
    display_name = attr.ib(type=str, factory=str)
    coach_email = attr.ib(type=str, factory=str)
    start = attr.ib(type=str, default=None)
    end = attr.ib(type=str, default=None)
    max_students_allowed = attr.ib(type=int, default=None)


@attr.s(frozen=True)
class CourseEnrollmentData:
    """
    Data related to a course enrollment object.

    This data is based on the fields available in the CourseEnrollment data model
    defined in the student app.

    Attributes:
        user (UserData): user associated with the Course Enrollment.
        course (CourseData): course where the user is enrolled in.
        mode (str): course mode associated with the course enrollment.
        is_active (bool): whether the enrollment is active.
        creation_date (datetime): creation date of the enrollment.
        created_by (UserData): if available, who created the enrollment.
    """

    user = attr.ib(type=UserData)
    course = attr.ib(type=CourseData)
    mode = attr.ib(type=str)
    is_active = attr.ib(type=bool)
    creation_date = attr.ib(type=datetime)
    created_by = attr.ib(type=UserData, default=None)


@attr.s(frozen=True)
class CertificateData:
    """
    Data related to a certificate object.

    This data is based on the GeneratedCertificate data model defined in the
    certificates app.

    Attributes:
        user (UserData): user associated with the Certificate.
        course (CourseData): course where the user obtained the certificate.
        mode (str): course mode associated with the course enrollment.
        grade (str): user's grade in this course run.
        download_url (str): URL where the PDF version of the certificate.
        name (str): user's name.
        current_status (str): current certificate status.
        previous_status (str): if available, pre-event certificate status.
    """

    user = attr.ib(type=UserData)
    course = attr.ib(type=CourseData)
    mode = attr.ib(type=str)
    grade = attr.ib(type=str)
    download_url = attr.ib(type=str)
    name = attr.ib(type=str)
    current_status = attr.ib(type=str)
    previous_status = attr.ib(type=str, factory=str)


@attr.s(frozen=True)
class CohortData:
    """
    Attributes defined for cohort membership object.

    This data is based on the fields available in the CohortMembership data model
    defined in the cohorts app.

    Attributes:
        user (UserData): user assigned to the group.
        course (CourseData): course associated with the course group.
        name (str): name of the cohort group.
    """

    user = attr.ib(type=UserData)
    course = attr.ib(type=CourseData)
    name = attr.ib(type=str)


@attr.s(frozen=True)
class DiscussionTopicContext:
    """
    Data related to a discussion topic context.

    Context for linking the external id for a discussion topic to its associated usage key.

    Attributes:
        title (str): title of the discussion. This field is cached to improve the performance, since otherwise we'd
          need to look it up in the course structure each time.
        usage_key (str): unit location.
        group_id (Optional[int]): can be used for providers that don't internally support
           cohorting but we can emulate that with different contexts for different groups.
        external_id (str): store the commentable id that is used by cs_comments_service.
        ordering (int): represent the position of the discussion topic.
        context (dict): additional structured information about the context in
           which this topic is used, such as the section, subsection etc.
    """

    title = attr.ib(type=str)
    usage_key = attr.ib(type=UsageKey, default=None)
    group_id = attr.ib(type=Optional[int], default=None)
    external_id = attr.ib(type=str, default=None)
    ordering = attr.ib(type=int, default=None)
    context = attr.ib(type=dict, factory=dict)


@attr.s(frozen=True)
class CourseDiscussionConfigurationData:
    """
    Data related to a course discussion configuration object.

    Course configuration information for discussions. Contains all the metadata
    needed to configure discussions for a course.

    Attributes:
        course_key (str): identifier of the course to which the discussion belongs.
        provider_type (str): provider type from discussion settings.
        enable_in_context (bool): indicates whether in-context discussion is enabled for the course
        enable_graded_units (bool): If enabled, discussion topics will be created for graded units as well.
        unit_level_visibility (bool): visibility for unit level.
        plugin_configuration (dict): The plugin configuration data for this context/provider.
        contexts (List[DiscussionTopicContext]): contains all the contexts for which discussion is to be enabled.
    """

    course_key = attr.ib(type=CourseKey)
    provider_type = attr.ib(type=str)
    enable_in_context = attr.ib(type=bool, default=True)
    enable_graded_units = attr.ib(type=bool, default=False)
    unit_level_visibility = attr.ib(type=bool, default=False)
    plugin_configuration = attr.ib(type=dict, default={})
    contexts = attr.ib(type=List[DiscussionTopicContext], factory=list)


@attr.s(frozen=True)
class PersistentCourseGradeData:
    """
    Data related to a persistent course grade object.

    This data is based on the fields available in the PersistentCourseGrade data model
    defined in the grades app.

    Attributes:
        user_id (int): identifier of the grade to which the grade belongs.
        course (CourseData): Identifier of the course to which the grade belongs.
        course_edited_timestamp (datetime): date the course was edited.
        course_version (str): version of the course.
        grading_policy_hash (str): grading policy hash of the course.
        percent_grade (float): percentage of the grade.
        letter_grade (str): grade in letter
        passed_timestamp (datetime): date the course was passed.
    """

    user_id = attr.ib(type=int)
    course = attr.ib(type=CourseData)
    course_edited_timestamp = attr.ib(type=datetime)
    course_version = attr.ib(type=str)
    grading_policy_hash = attr.ib(type=str)
    percent_grade = attr.ib(type=float)
    letter_grade = attr.ib(type=str)
    passed_timestamp = attr.ib(type=datetime)


@attr.s(frozen=True)
class XBlockSkillVerificationData:
    """
    Data needed to update verification count of tags/skills for an XBlock.

    User feedback on whether tags/skills related to an XBlock are valid.

    Attributes:
        usage_key (UsageKey): identifier of the XBlock object.
        verified_skills (List[int]): list of verified skill ids.
        ignored_skills (List[int]): list of ignored skill ids.
    """

    usage_key = attr.ib(type=UsageKey)
    verified_skills = attr.ib(type=List[int], factory=list)
    ignored_skills = attr.ib(type=List[int], factory=list)


@attr.s(frozen=True)
class UserNotificationData:
    """
    Data related to a user notification object.

    Attributes:
        user_ids (List(int)): identifier of the users to which the notification belongs.
        notification_type (str): type of the notification.
        content_url (str): url of the content.
        app_name (str): name of the app.
        course_key (str): identifier of the Course object.
        context (dict): additional structured information about the context of the notification.
    """

    user_ids = attr.ib(type=List[int])
    notification_type = attr.ib(type=str)
    content_url = attr.ib(type=str)
    app_name = attr.ib(type=str)
    course_key = attr.ib(type=CourseKey)
    context = attr.ib(type=dict, factory=dict)


@attr.s(frozen=True)
class ProgramData:
    """
    Data related to a program object.

    Attributes:
        uuid (str): The UUID of the program (from Course-Discovery).
        title (str): The title of the program.
        program_type (str): The type slug of the program (e.g. professional, microbachelors, micromasters, etc.).
    """

    uuid = attr.ib(type=str)
    title = attr.ib(type=str)
    program_type = attr.ib(type=str)


@attr.s(frozen=True)
class ProgramCertificateData:
    """
    Data related to a Program Certificate object.

    Attributes:
        user (UserData): User associated with the Program Certificate.
        program (ProgramData): Program data associated with the Program Certificate.
        uuid (str): UUID of the UserCredential record in Credentials.
        certificate_available_date (datetime): Optional. A DateTime describing when a learner is allowed to view the
           credential.
        status (str): The status of the credential (e.g. `awarded` or `revoked`).
        url (str): A URL to the learner's credential.
    """

    user = attr.ib(type=UserData)
    program = attr.ib(type=ProgramData)
    uuid = attr.ib(type=str)
    status = attr.ib(type=str)
    url = attr.ib(type=str)
    certificate_available_date = attr.ib(type=datetime, default=None)


@attr.s(frozen=True)
class ExamAttemptData:
    """
    Data for the Open edX Exam downstream effects.

    Note that events that use this data type:
        A. Pretain to "Special Exams", e.g. Timed or Proctored exams, and not non-timed course
        subsections that are labelled as an exam.
        B. Are only ever emitted from the newer exams backend, edx-exams, and never from the
        legacy exams backend, edx-proctoring.

    The event signals that use this data have the prefix `EXAM_`, which is equivalent to "special exam".
    We are using this as a shortened form of the prefix `SPECIAL_EXAM` to avoid confusion, as these are likely
    the only type of exams that will be of concern in the context of events/the event bus.

    For more information, please see the ADR relating to this event data:
    https://github.com/openedx/openedx-events/blob/main/docs/decisions/0013-special-exam-submission-and-review-events.rst

    Attributes:
        student_user (UserData): user object for the student to whom the exam attempt belongs
        course_key (CourseKey): identifier of the course to which the exam attempt belongs
        usage_key (UsageKey): identifier of the content that will get a exam attempt
        exam_type (str): type of exam that was taken (e.g. timed, proctored, etc.)
        requesting_user (UserData): user triggering the event (sometimes a non-learner, e.g. an instructor)
    """

    student_user = attr.ib(type=UserData)
    course_key = attr.ib(type=CourseKey)
    usage_key = attr.ib(type=UsageKey)
    exam_type = attr.ib(type=str)
    requesting_user = attr.ib(type=UserData, default=None)


@attr.s(frozen=True)
class CourseAccessRoleData:
    """
    Data related to a user's access role in a course.

    Attributes:
        user (UserData): user associated with the CourseAccessRole.
        course_key (CourseKey): identifier of the related course object.
        org (str): identifier of the organization.
        role (str): the role of the user in the course.
    """

    user = attr.ib(type=UserData)
    org_key = attr.ib(type=str)
    course_key = attr.ib(type=CourseKey)
    role = attr.ib(type=str)


@attr.s(frozen=True)
class DiscussionThreadData:
    """
    Data related to a discussion thread object, used to represent Forum events such as comments, responses, and threads.

    For more details on the data attributes, please see the following documentation:
    https://docs.openedx.org/en/latest/developers/references/internal_data_formats/tracking_logs/student_event_types.html#edx-forum-thread-created

    Attributes:
        anonymous (bool): indicates whether the user is anonymous.
        anonymous_to_peers (bool): indicates whether the user is anonymous to peers.
        body (str): body of the discussion thread.
        category_id (int): identifier of the category.
        category_name (str): name of the category.
        commentable_id (str): identifier of the commentable.
        group_id (int): identifier of the group.
        id (int): identifier of the discussion thread.
        team_id (int): identifier of the team.
        thread_type (str): type of the thread.
        title (str): title of the thread.
        title_truncated (bool): indicates whether the title is truncated.
        truncated (bool): indicates whether the thread is truncated.
        url (str): url of the thread.
        user (UserData): information of the user that authored the thread/comment/response.
        course_id (CourseKey): identifier of the course.
        discussion (dict): discussion data. (optional, specific to comments and responses)
        user_course_roles (List[str]): user course roles.
        user_forums_roles (List[str]): user forums roles.
        options (dict): options for the thread.
    """

    anonymous = attr.ib(type=bool)
    anonymous_to_peers = attr.ib(type=bool)
    body = attr.ib(type=str)
    category_id = attr.ib(type=int)
    category_name = attr.ib(type=str)
    commentable_id = attr.ib(type=str)
    group_id = attr.ib(type=int)
    id = attr.ib(type=int)
    team_id = attr.ib(type=int)
    thread_type = attr.ib(type=str)
    title = attr.ib(type=str)
    title_truncated = attr.ib(type=bool)
    truncated = attr.ib(type=bool)
    url = attr.ib(type=str)
    user = attr.ib(type=UserData)
    course_id = attr.ib(type=CourseKey)
    discussion = attr.ib(type=dict, factory=dict)
    user_course_roles = attr.ib(type=List[str], factory=list)
    user_forums_roles = attr.ib(type=List[str], factory=list)
    options = attr.ib(type=dict, factory=dict)


@attr.s(frozen=True)
class CourseNotificationData:
    """
    Data related to a course notification object.

    Attributes:
        course_key (str): identifier of the Course object.
        app_name (str): name of the app requesting the course notification.
        notification_type (str): type of the notification.
        content_url (str): url of the content the notification will redirect to.
        content_context (dict): additional information related to the content of the notification.
           Notification content templates are defined in edx-platform here:
           https://github.com/openedx/edx-platform/blob/master/openedx/core/djangoapps/notifications/base_notification.py#L10
        audience_filters (dict): additional information related to the audience of the notification.
           We can have different filters on course level, such as roles, enrollments, cohorts etc.

    Example of content_context for a discussion notification:

        >>> {
            ...,
            "content_context": {
                "post_title": "Post Title",
                "replier_name": "test_user",
        }


    Example of audience_filters for a discussion notification (new_discussion_post):

        >>> {
            ...,
            "audience_filters": {
                "enrollment": ["verified", "audit"],
                "role": ["discussion admin", "discussion moderator"],
        }
    """

    course_key = attr.ib(type=CourseKey)
    app_name = attr.ib(type=str)
    notification_type = attr.ib(type=str)
    content_url = attr.ib(type=str)
    content_context = attr.ib(type=dict, factory=dict)
    audience_filters = attr.ib(type=dict, factory=dict)


@attr.s(frozen=True)
class ORASubmissionAnswer:
    """
    Data related to the answer submitted by the user in an ORA submission.

    Attributes:
        parts (List[dict]): List with the response text in the ORA submission.

        The following attributes are used to represent the files submitted in the ORA submission:

        file_keys (List[str]): List of file keys in the ORA submission.
        file_descriptions (List[str]): List of file descriptions in the ORA submission.
        file_names (List[str]): List of file names in the ORA submission.
        file_sizes (List[int]): List of file sizes in the ORA submission.
        file_urls (List[str]): List of file URLs in the ORA submission.
    """

    parts = attr.ib(type=List[dict], factory=list)
    file_keys = attr.ib(type=List[str], factory=list)
    file_descriptions = attr.ib(type=List[str], factory=list)
    file_names = attr.ib(type=List[str], factory=list)
    file_sizes = attr.ib(type=List[int], factory=list)
    file_urls = attr.ib(type=List[str], factory=list)


@attr.s(frozen=True)
class ORASubmissionData:
    """
    Data associated to the ORA assessment submitted by a user.

    Attributes:
        uuid (str): The UUID of the ORA submission.
        anonymous_user_id (str): Optional. Anonymous user ID of the user who submitted the ORA submission.
        location (str): Optional. Location of the ORA submission.
        attempt_number (int): Attempt number of the ORA submission.
        created_at (datetime): Date and time when the ORA submission was created.
        submitted_at (datetime): Date and time when the ORA submission was submitted.
        answer (ORASubmissionAnswer): Answer submitted by the user in the ORA submission.
    """

    uuid = attr.ib(type=str)
    anonymous_user_id = attr.ib(type=str)
    location = attr.ib(type=str)
    attempt_number = attr.ib(type=int)
    created_at = attr.ib(type=datetime)
    submitted_at = attr.ib(type=datetime)
    answer = attr.ib(type=ORASubmissionAnswer)


@attr.s(frozen=True)
class CoursePassingStatusData:
    """
    Represents the event data when a user's grade is updated, indicates if current grade is enough for course passing.

    Attributes:
        is_passing (bool): Indicates whether the user's grade is enough to pass the course.
        user (UserData): An instance of UserData containing information about the user whose grade was updated.
        course (CourseData): An instance of CourseData containing details about the course
           in which the grade was updated.
    """

    is_passing = attr.ib(type=bool)
    course = attr.ib(type=CourseData)
    user = attr.ib(type=UserData)


@attr.s(frozen=True)
class CcxCoursePassingStatusData(CoursePassingStatusData):
    """
    Extends CoursePassingStatusData for CCX courses, specifying CCX course data.

    This class is used for events where a user's grade crosses a threshold specifically in a CCX course,
    providing a custom course attribute suited for CCX course instances.

    Attributes:
        course (CcxCourseData): An instance of CcxCourseData containing details about the CCX course
           in which the grade threshold was crossed.

        All other attributes are inherited from CoursePassingStatusData.
    """

    course = attr.ib(type=CcxCourseData)


@attr.s(frozen=True)
class BadgeTemplateData:
    """
    Data related to a badge template object.

    Attributes:
        uuid (str): UUID of the badge template.
        origin (str): type of badge template.
        name (str): badge name.
        description (str): badge description.
        image_url (str): badge image url.
    """

    uuid = attr.ib(type=str)
    origin = attr.ib(type=str)
    name = attr.ib(type=str, default=None)
    description = attr.ib(type=str, default=None)
    image_url = attr.ib(type=str, default=None)


@attr.s(frozen=True)
class BadgeData:
    """
    Data related to a badge object.

    Attributes:
        uuid (str): the UUID of the badge.
        user (UserData): user associated with the badge.
        template (BadgeTemplateData): badge template data.
    """

    uuid = attr.ib(type=str)
    user = attr.ib(type=UserData)
    template = attr.ib(type=BadgeTemplateData)


@attr.s(frozen=True)
class VerificationAttemptData:
    """
    Data related to a IDV attempt object.

    Attributes:
        attempt_id (int): the id of the verification attempt
        user (User): the user (usually a learner) performing the verification attempt.
        status (string): the status of the verification attempt.
        name (string): the name being ID verified. Defaults to None.
        expiration_datetime (datetime, optional): When the verification attempt expires. Defaults to None.
    """

    attempt_id = attr.ib(type=int)
    user = attr.ib(type=UserData)
    status = attr.ib(type=str)
    name = attr.ib(type=str, default=None)
    expiration_date = attr.ib(type=datetime, default=None)
